import configparser
from functools import partial
import os

import boto3

from serenata_toolbox.datasets.contextmanager import status_message
from serenata_toolbox.datasets.helpers import find_config


class RemoteDatasets:

    CONFIG = 'config.ini'

    def __init__(self):
        self.credentials = None
        self.client = None
        self.config = find_config(self.CONFIG)

        if not self.config_exists:
            print('Could not find {} file.'.format(self.CONFIG))
            print('You need Amazon section in it to interact with S3')
            print('(Check config.ini.example if you need a reference.)')
            return

        settings = configparser.RawConfigParser()
        settings.read(self.config)
        self.settings = partial(settings.get, 'Amazon')

        try:
            self.credentials = {
                'aws_access_key_id': self.settings('AccessKey'),
                'aws_secret_access_key': self.settings('SecretKey'),
                'region_name': self.settings('Region')
            }

            # friendly user message warning about old config.ini version
            region = self.credentials.get('region_name', '')
            if region and region.startswith('s3-'):
                msg = (
                    'It looks like you have an old version of the config.ini '
                    'file. We do not need anymore the service (s3) appended '
                    'to the region (sa-east-1). Please update your config.ini '
                    'replacing regions like `s3-sa-east-1` by `sa-east-1`.'
                )
                print(msg)

        except configparser.NoSectionError:
            msg = (
                'You need an Amazon section in {} to interact with S3 '
                '(Check config.ini.example if you need a reference.)'
            )
            print(msg.format(self.CONFIG))

    @property
    def config_exists(self):
        return all((os.path.exists(self.config), os.path.isfile(self.config)))

    @property
    def bucket(self):
        if hasattr(self, 'settings'):
            try:
                return self.settings('Bucket')
            except configparser.NoSectionError:
                return None

    @property
    def s3(self):
        if not self.client and self.credentials:
            self.client = boto3.client('s3', **self.credentials)
        return self.client

    @property
    def all(self):
        if self.s3 and self.bucket:
            response = self.s3.list_objects(Bucket=self.bucket)
            yield from (obj.get('Key') for obj in response.get('Contents', []))

    def upload(self, file_path):
        if self.s3 and self.bucket:
            _, file_name = os.path.split(file_path)
            with status_message('Uploading {}…'.format(file_name)):
                self.s3.upload_file(file_path, self.bucket, file_name)

    def delete(self, file_name):
        if self.s3 and self.bucket:
            with status_message('Deleting {}…'.format(file_name)):
                self.s3.delete_object(Bucket=self.bucket, Key=file_name)



import os

from serenata_toolbox.datasets.downloader import Downloader
from serenata_toolbox.datasets.local import LocalDatasets
from serenata_toolbox.datasets.remote import RemoteDatasets


class Datasets:
    """
    This is a wrapper for three different classes that together handle the
    datasets (locally and remotely).

    Datasets class takes one argument: the path to the local directory of the
    dataset files (e.g. data/ or /tmp/serenata-data). The argument is optional
    and the default value is data/ (following the default usage in the main
    repo, serenata-de-amor).

    The remote part of the class expect to find Amazon credentials in a
    config.ini file with an Amazon section (e.g. config.ini.exemple).

    Inside it object there are three main objects: local, remote, and
    downloader:

    * `Datasets.local` handles listing all local datasets through the property
      `Datasets.local.all` (hint: it's a generator) and deleting local datasets
      with the method `Datasets.local.delete(filename)`;

    * `Datasets.remote` has the `Datasets.remote.all` property (hint: it's also
      a generator) and `Dataset.remote.delete(filename)` method just like its
      local equivalent; in addition to them this object offers the
      `Datasets.remote.upload(file_path)` method to upload a local file to the
      remote bucket; `Datasets.remote` does not handles downloads because
      `boto3` does not support `asyncio` and we prefer to use async tasks to
      allow the download of more than one file in parallel;

    * `Datasets.downloader` implements a async manager to download files from
      the remote bucket. It's `Datasets.downloader.download(files)` take the
      path for a single file (str) as argument or an iterable of paths (str).

    Yet this wrapper implement the `Dataset.upload_all()` method to upload all
    local datasets that are not present in the remote bucket.

    :param local_directory: (str) path to local directory of the datasets
    """

    def __init__(self, local_directory=None):
        if not local_directory:
            local_directory = 'data'

        self.local = LocalDatasets(local_directory)
        self.remote = RemoteDatasets()
        self.downloader = Downloader(
            local_directory,
            bucket=self.remote.bucket,
            **self.remote.credentials
        )

    @property
    def pending(self):
        """Files that are in the local datasets but not in S3."""
        local = set(self.local.all)
        remote = set(self.remote.all)
        yield from (local - remote)

    def upload_all(self):
        for file_name in self.pending:
            full_path = os.path.join(self.local.directory, file_name)
            self.remote.upload(full_path)


# shortcuts & retrocompatibility

def fetch(filename, destination_path):
    datasets = Datasets(destination_path)
    return datasets.downloader.download(filename)


def fetch_latest_backup(destination_path, force_all=False):
    datasets = Datasets(destination_path)

    if force_all:
        files = datasets.downloader.LATEST
    else:
        files = tuple(
            f for f in datasets.downloader.LATEST
            if not os.path.exists(os.path.join(destination_path, f))
        )

    if not files:
        print('You already have all the latest datasets! Nothing to download.')

    return datasets.downloader.download(files)



import os

from datetime import datetime

CSV_PARAMS = {
    'compression': 'xz',
    'encoding': 'utf-8',
    'index': False
}


# Utilities for extracting data from XML files

def xml_extract_text(node, xpath):
    """
    :param node: the node to be queried
    :param xpath: the path to fetch the child node that has the wanted text
    """
    text = node.find(xpath).text
    if text is not None:
        text = text.strip()
    return text


def xml_extract_date(node, xpath, date_format='%d/%m/%Y'):
    """
    :param node: the node to be queried
    :param xpath: the path to fetch the child node that has the wanted date
    """
    return datetime.strptime(xml_extract_text(node, xpath), date_format)


def xml_extract_datetime(node, xpath, datetime_format='%d/%m/%Y %H:%M:%S'):
    """
    :param node: the node to be queried
    :param xpath: the path to fetch the child node that has the wanted datetime
    """
    return datetime.strptime(xml_extract_text(node, xpath), datetime_format)


# Utilities for dealing with Dataframes

def translate_column(df, column, translations):
    """
    :param df: (pandas.Dataframe) the dataframe to be translated
    :param column: (str) the column to be translated
    :param translations: (dict) a dictionary of the strings to be categorized and translated
    """
    df[column] = df[column].astype('category')
    translations = [translations[cat]
                    for cat in df[column].cat.categories]
    df[column].cat.rename_categories(translations, inplace=True)


def save_to_csv(df, data_dir, name):
    today = datetime.strftime(datetime.now(), '%Y-%m-%d')
    file_path = os.path.join(data_dir, '{}-{}.xz'.format(today, name))
    df.to_csv(file_path, **CSV_PARAMS)


# Utility for config file lookup (credits to fabric fabfile lookup)

def find_config(name='config.ini'):
    """
    :param name: config file name (defaults to 'config.ini')
    """
    # start in cwd and work downwards towards filesystem root
    path = '.'
    # Stop before falling off root of filesystem (should be platform
    # agnostic)
    while os.path.split(os.path.abspath(path))[1]:
        joined = os.path.join(path, name)

        # return config file absolute path if exists
        if os.path.exists(joined) and os.path.isfile(joined):
            return os.path.abspath(joined)

        path = os.path.join('..', path)

    # if config not found fallback on tests expected behavior
    # wich is to return the filename to be used at remote
    # config_exists class property.
    # a future refactor could be to return None and not use
    # config exists anymore
    return name



import csv
import os.path
from urllib.request import urlretrieve
from zipfile import ZipFile

import numpy as np
import pandas as pd

from .reimbursements import AVAILABLE_YEARS, Reimbursements


class Dataset:
    def __init__(self, path, years=AVAILABLE_YEARS):
        self.path = path
        self.years = years if isinstance(years, list) else [years]
        self.subquotas = (
            (1, 'Maintenance of office supporting parliamentary activity'),
            (2, 'Locomotion, meal and lodging'),
            (3, 'Fuels and lubricants'),
            (4, 'Consultancy, research and technical work'),
            (5, 'Publicity of parliamentary activity'),
            (6, 'Purchase of office supplies'),
            (7, 'Software purchase or renting; Postal services; Subscriptions'),
            (8, 'Security service provided by specialized company'),
            (9, 'Flight tickets'),
            (10, 'Telecommunication'),
            (11, 'Postal services'),
            (12, 'Publication subscriptions'),
            (13, 'Congressperson meal'),
            (14, 'Lodging, except for congressperson from Distrito Federal'),
            (15, 'Automotive vehicle renting or watercraft charter'),
            (119, 'Aircraft renting or charter of aircraft'),
            (120, 'Automotive vehicle renting or charter'),
            (121, 'Watercraft renting or charter'),
            (122, 'Taxi, toll and parking'),
            (123, 'Terrestrial, maritime and fluvial tickets'),
            (137, 'Participation in course, talk or similar event'),
            (999, 'Flight ticket issue')
        )

        self.translate_columns = {
            'ideDocumento': 'document_id',
            'txNomeParlamentar': 'congressperson_name',
            'idecadastro': 'congressperson_id',
            'nuCarteiraParlamentar': 'congressperson_document',
            'nuLegislatura': 'term',
            'sgUF': 'state',
            'sgPartido': 'party',
            'codLegislatura': 'term_id',
            'numSubCota': 'subquota_number',
            'txtDescricao': 'subquota_description',
            'numEspecificacaoSubCota': 'subquota_group_id',
            'txtDescricaoEspecificacao': 'subquota_group_description',
            'txtFornecedor': 'supplier',
            'txtCNPJCPF': 'cnpj_cpf',
            'txtNumero': 'document_number',
            'indTipoDocumento': 'document_type',
            'datEmissao': 'issue_date',
            'vlrDocumento': 'document_value',
            'vlrGlosa': 'remark_value',
            'vlrLiquido': 'net_value',
            'numMes': 'month',
            'numAno': 'year',
            'numParcela': 'installment',
            'txtPassageiro': 'passenger',
            'txtTrecho': 'leg_of_the_trip',
            'numLote': 'batch_number',
            'numRessarcimento': 'reimbursement_number',
            'vlrRestituicao': 'reimbursement_value',
            'nuDeputadoId': 'applicant_id',
        }

    def fetch(self):
        base_url = "http://www.camara.leg.br/cotas/Ano-{}.csv.zip"
        retrieved_files = []

        for year in self.years:
            zip_file_path = os.path.join(self.path, "Ano-{}.zip".format(year))
            url = base_url.format(year)
            urlretrieve(url, zip_file_path)
            retrieved_files.append(zip_file_path)

            self._extract_zip_file(zip_file_path)
            os.remove(zip_file_path)

        datasets_format = 'datasets-format.html'
        url = ('http://www2.camara.leg.br/transparencia/cota-para-exercicio-da-atividade-parlamentar/'
               'explicacoes-sobre-o-formato-dos-arquivos-xml')
        urlretrieve(url, os.path.join(self.path, datasets_format))
        retrieved_files.append(datasets_format)

        return retrieved_files

    def _extract_zip_file(self, zip_file_path):
        zip_file = ZipFile(zip_file_path, 'r')
        zip_file.extractall(self.path)
        zip_file.close()

        return True

    def translate(self):
        for year in self.years:
            csv_path = os.path.join(self.path, 'Ano-{}.csv'.format(year))
            self._translate_file(csv_path)

    def clean(self):
        reimbursements = Reimbursements(self.path, years=self.years)
        dataset = reimbursements.group(reimbursements.receipts)
        reimbursements.write_reimbursement_file(dataset)

    def _translate_file(self, csv_path):
        output_file_path = csv_path \
            .replace('.csv', '.xz') \
            .replace('Ano-', 'reimbursements-')

        data = pd.read_csv(csv_path,
                           encoding='utf-8',
                           delimiter=";",
                           quoting=csv.QUOTE_NONE,
                           decimal=",",
                           dtype={'ideDocumento': np.str,
                                  'idecadastro': np.str,
                                  'nuCarteiraParlamentar': np.str,
                                  'codLegislatura': np.str,
                                  'txtCNPJCPF': np.str,
                                  'numRessarcimento': np.str},
                           )

        data.rename(columns=self.translate_columns, inplace=True)

        for code, name in self.subquotas:
            data.loc[data['subquota_number'] == code, ['subquota_description']] = name

        data.to_csv(output_file_path, compression='xz', index=False,
                    encoding='utf-8')

        return output_file_path



import os

from datetime import date

import pandas as pd
import numpy as np

AVAILABLE_YEARS = [year for year in range(2009, date.today().year + 1)]


class Reimbursements:

    FILE_BASE_NAME = 'reimbursements.xz'

    CSV_PARAMS = {
        'compression': 'xz',
        'encoding': 'utf-8',
        'index': False
    }

    def __init__(self, path, years=AVAILABLE_YEARS):
        self.path = path
        self.years = years if isinstance(years, list) else [years]

    def read_csv(self, name):
        filepath = os.path.join(self.path, name)
        print('Loading {}…'.format(name))
        dtype = {
            'applicant_id': np.str,
            'batch_number': np.str,
            'cnpj_cpf': np.str,
            'congressperson_document': np.str,
            'congressperson_id': np.str,
            'document_id': np.str,
            'document_number': np.str,
            'document_type': np.str,
            'leg_of_the_trip': np.str,
            'passenger': np.str,
            'reimbursement_number': np.str,
            'subquota_group_description': np.str,
            'subquota_group_id': np.str,
            'subquota_number': np.str,
            'term_id': np.str,
        }
        return pd.read_csv(filepath, dtype=dtype)

    @property
    def receipts(self):
        print('Merging all datasets…')
        datasets = ["reimbursements-{}.xz".format(n) for n in self.years]
        data = (self.read_csv(name) for name in datasets)
        return pd.concat(data)

    @staticmethod
    def aggregate(grouped, old, new, func):
        """
        Gets a GroupBy object, aggregates it on `old` using `func`, then rename
        the series name from `old` to `new`, returning a DataFrame.
        """
        output = grouped[old].agg(func)
        output = output.rename(index=new, inplace=True)
        return output.reset_index()

    def group(self, receipts):
        print('Dropping rows without document_value or reimbursement_number…')
        subset = ('document_value', 'reimbursement_number')
        receipts = receipts.dropna(subset=subset)

        groupby_keys = ('year', 'applicant_id', 'document_id')
        receipts = receipts.dropna(subset=subset + groupby_keys)

        receipts = receipts[receipts['document_value'] != 0]
        receipts = receipts[receipts['reimbursement_number'] != '0']
        receipts = receipts[receipts['year'] != 0]
        receipts = receipts[receipts['applicant_id'] != '0']
        receipts = receipts[receipts['document_id'] != '0']

        print('Grouping dataset by applicant_id, document_id and year…')
        grouped = receipts.groupby(groupby_keys)

        print('Gathering all reimbursement numbers together…')
        numbers = self.aggregate(
            grouped,
            'reimbursement_number',
            'reimbursement_numbers',
            lambda x: ', '.join(set(x))
        )

        print('Summing all net values together…')
        net_total = self.aggregate(
            grouped,
            'net_value',
            'total_net_value',
            np.sum
        )

        print('Summing all reimbursement values together…')
        total = self.aggregate(
            grouped,
            'reimbursement_value',
            'reimbursement_value_total',
            np.sum
        )

        print('Generating the new dataset…')
        final = pd.merge(
            pd.merge(pd.merge(total, net_total, on=groupby_keys), numbers, on=groupby_keys),
            receipts,
            on=groupby_keys
        )
        final = final.drop_duplicates(subset=groupby_keys)
        final.rename(columns={'net_value': 'net_values',
                              'reimbursement_value': 'reimbursement_values'},
                     inplace=True)
        final = final.drop('reimbursement_number', 1)
        return final

    @staticmethod
    def unique_str(strings):
        return ', '.join(set(strings))

    def write_reimbursement_file(self, receipts):
        print('Casting changes to a new DataFrame…')
        df = pd.DataFrame(data=receipts)

        print('Writing it to file…')
        filepath = os.path.join(self.path, self.FILE_BASE_NAME)
        df.to_csv(filepath, **self.CSV_PARAMS)

        print('Done.')
